/****************************************************************************
**
** Copyright (C) 2006-2008 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 2 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "automaton.h"

#include <QStack>
#include <QDebug>

static void addPattern(
				Automaton *a,
				const QString& pattern, bool caseSensitive,
				int i, int end,
				StateId forkBeg, StateId forkEnd,
				quint32 format = 0)
{
	if ( end < 0 )
		end = pattern.length() + end + 1;
	else
		end = qMin(end, pattern.length());
	
	//QString rpattern;
	//
	//for ( int j = i; j < end; ++j )
	//	rpattern += pattern.at(j);
	//
	//qDebug("addPattern(\"%s\");", qPrintable(pattern.mid(i, end - i)));
	
	--i;
	
	StateId previous = forkBeg, current = forkBeg, fork = 0xffffffff;
	
	while ( ++i < end )
	{
		QChar c = pattern.at(i);
		
		if ( c == QLatin1Char('(') )
		{
			int idx = i, nest = 1;
			
			while ( nest && (++idx < end) )
			{
				if ( pattern.at(idx) == QLatin1Char('\\') )
				{
					++idx;
					continue;
				} else if ( pattern.at(idx) == QLatin1Char('(') ) {
					++nest;
				} else if ( pattern.at(idx) == QLatin1Char(')') ) {
					--nest;
				} else if ( pattern.at(idx) == QLatin1Char('[') ) {
					while ( ++idx < end )
					{
						if ( pattern.at(idx) == QLatin1Char('\\') )
						{
							++idx;
							continue;
						} else if ( pattern.at(idx) == QLatin1Char(']') ) {
							break;
						}
					}
				}
			}
			
			if ( (idx >= end) || nest )
			{
				qWarning("Parsing error : supplied regexp is syntactically invalid.\n[%s]\nIrregularity from pos %i.", qPrintable(pattern), i);
				return;
			}
			
			StateId sbeg = a->createState(),
					send = a->createState();
			
			a->addTransition(current, Epsilon, sbeg);
			
			addPattern(a, pattern, caseSensitive, i + 1, idx, sbeg, send);
			
			previous = current;
			current = send;
			fork = sbeg;
			i = idx;
			
			continue;
			
		} else if ( c == QLatin1Char('{') ) {
			// new operator : recursion...
			
			int idx = i, nest = 1, blocknest = 0, mid = -1;
			
			while ( nest && (++idx < end) )
			{
				quint16 c = pattern.at(idx).unicode();
				
				if ( c == '\\' )
				{
					++idx;
					continue;
				} else if ( c == '{' ) {
					++nest;
				} else if ( c == '}' ) {
					--nest;
				} else if ( c == '(' ) {
					++blocknest;
				} else if ( c == ')' ) {
					--blocknest;
				} else if ( c == '[' ) {
					while ( ++idx < end )
					{
						if ( pattern.at(idx) == QLatin1Char('\\') )
						{
							++idx;
							continue;
						} else if ( pattern.at(idx) == QLatin1Char(']') ) {
							break;
						}
					}
				} else if ( nest == 1 && c == '.' ) {
					if ( mid != -1 )
						qWarning("Invalid context operator");
					
					mid = idx;
				}
			}
			
			if ( (idx >= end) || nest )
			{
				qWarning("Parsing error : supplied regexp is syntactically invalid.\n[%s]\nIrregularity from pos %i.", qPrintable(pattern), i);
				return;
			}
			
			StateId sbeg = a->createState(),
					send = a->createState();
			
			a->addTransition(current, Epsilon, sbeg);
			
			addPattern(a, pattern, caseSensitive, i + 1, mid, sbeg, send, format);
			
			// add match for context begin
			a->match[send] = qMatch(Context | Begin, a->sub.count(), 1); //format);
			
			// add new context
			Automaton *sub = new Automaton;
			sub->parent = a;
			sub->format = format;
			a->sub << sub;
			
			// add context end
			StateId lend = sub->createState(),
					lrec = sub->createState();
			
			addPattern(sub, pattern, caseSensitive, mid + 1, idx, 0, lend, format);
			sub->match[lend] = qMatch(Context | End, 0, 1);
			
			// recursion trick
			addPattern(sub, pattern, caseSensitive, i + 1, mid, 0, lrec, format);
			sub->match[lrec] = qMatch(Context | Begin, sub->sub.count(), 1); //format);
			sub->sub << sub;
			
			previous = current;
			current = send;
			fork = sbeg;
			i = idx;
			
			continue;
			
		} else if ( c == QLatin1Char('|') ) {
			//fork = -1;
			
			a->addTransition(current, Epsilon, forkEnd);
			
			fork = previous = current = forkBeg;
			continue;
		} else if ( c == QLatin1Char('[') ) {
			
			StateId sbeg = a->createState(),
					send = a->createState();
			
			a->addTransition(current, Epsilon, sbeg);
			
			if ( ((i+1) < end) && (pattern.at(i + 1) == QLatin1Char('^')) )
			{
				++i;
				a->match[send] = qMatch(BlackHole, 0, 0);
			}
			
			while ( ++i < end )
			{
				if ( pattern.at(i) == QLatin1Char('\\') )
				{
					if ( ++i >= end )
						break;
					
				} else if ( pattern.at(i) == QLatin1Char(']') ) {
					break;
				} else if ( pattern.at(i) == QLatin1Char('-') ) {
					
					quint16 start = pattern.at(i - 1).unicode() + 1, end = pattern.at(i + 1).unicode();
					
					for ( quint16 sc = start; sc <= end; ++sc )
						a->addTransition(sbeg, sc, send);
					
					++i;
				}
				
				a->addTransition(sbeg, pattern.at(i).unicode(), send);
			}
			
			previous = current;
			current = send;
			fork = sbeg;
			
			continue;
		} else if ( (c == QLatin1Char('\\')) && ((i+1) < end) ) {
			c = pattern.at(++i);
			
			if ( c == 'n' )
				c = '\n';
			else if ( c == 't' )
				c = '\t';
			else if ( c == 'v' )
				c = '\v';
			else if ( c == 'a' )
				c = '\a';
		} else if ( c == QLatin1Char('*') ) {
			
			a->addTransition(current, Epsilon, fork);
			//a->addTransition(current, Epsilon, (fork == 0xffffffff) ? current : fork);
			
			StateId next = a->createState();
			
			a->addTransition(current, Epsilon, next);
			a->addTransition(fork, Epsilon, next);
			//a->addTransition((fork == 0xffffffff) ? previous : fork, Epsilon, next);
			current = next;
			
			continue;
		} else if ( c == QLatin1Char('+') ) {
			// simple auto loop...
			
			//a->addTransition(current, Epsilon, (fork == 0xffffffff) ? current : fork);
			a->addTransition(current, Epsilon, fork);
			
			continue;
		} else if ( c == QLatin1Char('?') ) {
			
			StateId next = a->createState();
			
			a->addTransition(current, Epsilon, next);
			//a->addTransition((fork == 0xffffffff) ? previous : fork, Epsilon, next);
			a->addTransition(fork, Epsilon, next);
			current = next;
			
			continue;
		}
		
		previous = current;
		
		fork = a->createState();
		a->addTransition(current, Epsilon, fork);
		
		current = a->createState();
		
		if ( !caseSensitive && c.isLetter() )
		{
			StateId base = a->createState();
			
			a->addTransition(fork, c.toLower().unicode(), current);
			a->addTransition(fork, c.toUpper().unicode(), current);
		} else {
			a->addTransition(fork, c.unicode(), current);
		}
	}
	
	a->addTransition(current, Epsilon, forkEnd);
}

Automaton::Automaton()
 : det(false), cref(0), format(0), flags(None), parent(0)
{
	ref();
	
	//qDebug("creating 0x%lx", this);
	
	createStandaloneState();
}

Automaton::~Automaton()
{
	if ( parent )
		parent->sub.removeAll(this);
	
	parent = 0;
	
	//qDeleteAll(sub);
	foreach ( Automaton *c, sub )
		c->deref();
}

void Automaton::ref()
{
	++cref;
}

void Automaton::deref()
{
	--cref;
	
	if ( cref == 0 )
		delete this;
}

Automaton* Automaton::clone(quint32 formatOffset) const
{
	Automaton *a = new Automaton;
	a->id = id;
	a->det = det;
	a->flags = flags;
	a->format = format + formatOffset;
	
	a->edges = edges;
	a->trans = trans;
	
	a->match = match;
	
	a->standalone = standalone;
	
	QHash<StateId, Match>::iterator it = a->match.begin();
	
	while ( it != a->match.end() )
	{
		it->format += formatOffset;
		
		++it;
	}
	
	foreach ( Automaton *c, sub )
		if ( c != this )
			a->sub << c->clone(formatOffset);
		else
			a->sub << a;
	
	return a;
}

StateId Automaton::createState()
{
	StateId s = trans.count();
	
	trans << TransitionList();
	edges << StateList();
	
	return s;
}

StateId Automaton::createStandaloneState()
{
	StateId s = trans.count();
	
	trans << TransitionList();
	edges << StateList();
	
	standalone << s;
	
	//qDebug("creating standalone state %i", s);
	
	return s;
}

void Automaton::addTransition(StateId from, Transition t, StateId to)
{
	trans[from] << t;
	edges[from] << to;
}

StateId Automaton::addWord(const QString& s, bool caseSensitive, Match m)
{
	//qDebug("// word : %s", qPrintable(s));
	//::addPattern(this, s, caseSensitive, 0, s.length(), start.id, match.id);
	StateId current = 0;
	
	for ( int i = 0; i < s.length(); ++i )
	{
		if ( s.at(i).isLetter() && !caseSensitive )
		{
			StateId sid = createState();
			
			addTransition(current, s.at(i).toLower().unicode(), sid);
			addTransition(current, s.at(i).toUpper().unicode(), sid);
			
			current = sid;
		} else {
			StateId sid = createState();
			
			addTransition(current, s.at(i).unicode(), sid);
			
			current = sid;
		}
	}
	
	match.insert(current, m);
	
	return current;
}

StateId Automaton::addPattern(const QString& s, bool caseSensitive, Match m)
{
	//qDebug("// pattern : %s", qPrintable(s));
	StateId sm = createState();
	
	::addPattern(this, s, caseSensitive, 0, s.length(), 0, sm, qMatchFormat(m));
	
	match.insert(sm, m);
	
	return sm;
}

void Automaton::addAutomatonContents(Automaton *a, quint32 formatOffset)
{
	insertAutomatonContents(a, 0, 0, formatOffset);
	
	/*
	int max = qMin(a->edges.count(), a->trans.count());
	
	int offset = edges.count() - 1;
	int stdoff = standalone.count();
	
	for ( int i = 1; i < a->standalone.count(); ++i )
		standalone << a->standalone[i] + offset;
	
	// start state : special placement
	const StateList& isl = a->edges.at(0);
	
	for ( int eid = 0; eid < isl.count(); ++eid )
	{
		edges[0] << isl.at(eid) + offset;
	}
	
	trans[0] << a->trans.at(0);
	
	//qDebug("adding %i states [%i, %i]", max - 1, offset + 1, offset + max - 1);
	
	// add other states / transitions
	for ( int sid = 1; sid < max; ++sid )
	{
		StateList nsl;
		const StateList& sl = a->edges.at(sid);
		
		for ( int eid = 0; eid < sl.count(); ++eid )
			nsl << sl.at(eid) + offset;
		
		trans << a->trans.at(sid);
		edges << nsl;
	}
	
	// match information
	QList<Automaton*> children;
	QHash<StateId, Match>::const_iterator it = a->match.constBegin();
	
	while ( it != a->match.constEnd() )
	{
		Match m = *it;
		quint32 jmp = qMatchJump(m);
		
		if ( (m.type & Context) && (m.type & Begin) )
		{
			if ( m.type & BlackHole )
			{
				quint32 njmp = jmp + stdoff - 1;
				
				m.type = qMatchType(m) | njmp;
				
			} else {
				Automaton *c = a->sub.at(jmp);
				
				if ( children.contains(c) )
				{
					m.type = qMatchType(m) | (children.indexOf(c) + sub.count());
				} else {
					m.type = qMatchType(m) | (children.count() + sub.count());
					children << c;
				}
			}
			//qDebug("jump : %i", qMatchJump(m));
		}
		
		m.format += formatOffset;
		
		//qDebug("format update : %i", m.format);
		
		match.insert(it.key() + offset, m);
		
		++it;
	}
	
	foreach ( Automaton *c, a->children )
	{
		sub << c->clone(formatOffset);
		//qDebug("format update : %i", c->format);
	}
	*/
}

void Automaton::insertAutomatonContents(Automaton *a, StateId start, StateId end, quint32 formatOffset)
{
	int max = qMin(a->edges.count(), a->trans.count());
	
	int offset = edges.count();
	int stdoff = standalone.count();
	
	//qDebug("(0x%x)inserting %i std states", this, a->standalone.count() - 1);
	
	for ( int i = 1; i < a->standalone.count(); ++i )
	{
		//qDebug("added std state %i (%i becomes %i)", standalone.count(), a->standalone[i], a->standalone[i] + offset);
		standalone << a->standalone[i] + offset;
	}
	
	// add other states / transitions
	for ( int sid = 0; sid < max; ++sid )
	{
		StateList nsl;
		const StateList& sl = a->edges.at(sid);
		
		for ( int eid = 0; eid < sl.count(); ++eid )
			nsl << sl.at(eid) + offset;
		
		trans << a->trans.at(sid);
		edges << nsl;
	}
	
	// connect
	addTransition(start, Epsilon, offset);
	
	// match information
	QList<Automaton*> children;
	QHash<StateId, Match>::const_iterator it = a->match.constBegin();
	
	while ( it != a->match.constEnd() )
	{
		Match m = *it;
		quint32 jmp = qMatchJump(m);
		
		if ( m.type & Context )
		{
			if ( m.type & Begin )
			{
				Automaton *c = a->sub.at(jmp);
				
				if ( children.contains(c) )
				{
					m.type = qMatchType(m) | (children.indexOf(c) + sub.count());
				} else {
					m.type = qMatchType(m) | (children.count() + sub.count());
					children << c;
				}
			} else if ( m.type & BlackHole ) {
				quint32 njmp = jmp + stdoff - 1;
				
				m.type = qMatchType(m) | njmp;
				
				//standalone[njmp] = a->standalone[jmp] + offset;
				
				//qDebug("jump[0] : %i -> %i", jmp, njmp);
			}
			//qDebug("jump : %i", qMatchJump(m));
		} else if ( m.type & Anchor ) {
			
			Automaton *c = a->sub.at(jmp);
			
			if ( children.contains(c) )
			{
				m.type = qMatchType(m) | (children.indexOf(c) + sub.count());
			} else {
				m.type = qMatchType(m) | (children.count() + sub.count());
				children << c;
			}
		} else if ( end ) {
			// connect
			
			addTransition(it.key() + offset, Epsilon, end);
			++it;
			continue;
		}
		
		m.format += formatOffset;
		
		//qDebug("format update : %i", m.format);
		
		match.insertMulti(it.key() + offset, m);
		
		++it;
	}
	
	foreach ( Automaton *c, children )
	{
		Automaton *cc = 0;
		
		if ( c == a || c == this )
		{
			sub << this;
			
			cc = (c == a) ? c : 0;
		} else {
			cc = c->clone(formatOffset);
			sub << cc;
		}
		
		if ( !cc )
			continue;
		
		QHash<StateId, Match>::iterator it = cc->match.begin();
		
		while ( it != cc->match.end() )
		{
			Match m = *it;
			quint32 jmp = qMatchJump(m);
			
			if ( qMatchType(m) == (Context | End | BlackHole) )
			{
				quint32 njmp = jmp + stdoff - 1;
				
				it->type = qMatchType(m) | njmp;
				
				//standalone[njmp] = a->standalone[jmp] + offset;
				
				//qDebug("jump[0x%x] : %i -> %i", cc, jmp, njmp);
			}
			
			++it;
		}
		
		//qDebug("format update : %i", c->format);
	}
}

StateList Automaton::move(StateList l, Transition t) const
{
	StateList dest;
	
	foreach ( StateId sid, l )
	{
		const StateList& sl = edges.at(sid);
		const TransitionList& tl = trans.at(sid);
		
		for ( int tid = 0; tid < tl.count(); ++tid )
		{
			if ( tl.at(tid) == t )
			{
				StateId dsid = sl.at(tid);
				
				if ( !dest.contains(dsid) )
					dest << dsid;
			}
		}
	}
	
	return dest;
}

StateList Automaton::epsilonClosure(StateId s) const
{
	StateList walked;
	
	walked << s;
	int walk = -1;
	
	while ( ++walk < walked.count() )
	{
		StateId sid = walked.at(walk);
		const StateList& sl = edges.at(sid);
		const TransitionList& tl = trans.at(sid);
		
		for ( int i = 0; i < tl.count(); ++i )
		{
			if ( tl.at(i) == Epsilon )
			{
				StateId es = sl.at(i);
				
				if ( !walked.contains(es) )
					walked << es;
			}
		}
	}
	
	return walked;
}

StateList Automaton::epsilonClosure(const StateList& l) const
{
	StateList ec = l;
	QStack<StateId> stack;
	
	for ( int i = 0; i < l.count(); ++i )
	{
		StateList tmp = epsilonClosure(l.at(i));
		
		for ( quint32 j = 0; j < tmp.count(); ++j )
		{
			StateId es = tmp.at(j);
			
			if ( !ec.contains(es) )
				ec << es;
		}
	}
	
	return ec;
}

Automaton* Automaton::determinize() const
{
	Automaton *d = new Automaton;
	d->id = id;
	d->det = true;
	d->flags = flags;
	d->format = format;
	
	int index = -1;
	QList<StateList> dstates;
	bool lastStandaloneFiller = false;
	
	d->trans.clear();
	d->edges.clear();
	d->standalone.clear();
	
	foreach ( StateId root, standalone )
	{
		if ( root == 0xffffffff )
		{
			// special filler state needed for recursive subcontexts
			d->standalone << 0xffffffff;
			lastStandaloneFiller = true;
			continue;
		} else if ( lastStandaloneFiller ) {
			if ( root == 0 )
				d->standalone << 0;
			else
				qWarning("unsupported use of std:filler");
			
			lastStandaloneFiller = false;
		}
		
		StateId nroot = d->createStandaloneState();
		
		StateList l = epsilonClosure(root);
		
		if ( !l.contains(root) )
			l << root;
		
		// sort states equiv to be able to use QList::operator == ()
		qSort(l);
		
		dstates << l;
		
		while ( ++index < dstates.count() )
		{
			//qDebug() << "filling trans table for" << l;
			
			l = dstates.at(index);
			
			// compute local alphabet for DFA state
			QList<Transition> localAlphabet;
			
			foreach ( StateId sid, l )
			{
				const TransitionList& lt = trans.at(sid);
				
				foreach ( Transition t, lt )
				{
					if ( (t != Epsilon) && !localAlphabet.contains(t) )
						localAlphabet << t;
				}
			}
			
			// sort alphabet to place Any at end
			qSort(localAlphabet);
			
			foreach ( Transition t, localAlphabet )
			{
				StateList cedges = epsilonClosure(move(l, t));
				
				qSort(cedges);
				
				int idx = dstates.indexOf(cedges);
				
				if ( idx == -1 )
				{
					// new state
					StateId ns = d->createState();
					
					if ( ns != (quint32)dstates.count() )
						qWarning("NFA=>DFA reduction error...");
					
					dstates << cedges;
					
					idx = ns;
				}
				
				d->addTransition(index, t, idx);
			}
		}
	}
	
	// backport matches
	
	for ( int i = 0; i < dstates.count(); ++i )
	{
		//
		const StateList& l = dstates.at(i);
		
		foreach ( StateId sid, l )
		{
			QHash<StateId, Match>::const_iterator it = match.constFind(sid);
			
			while ( it != match.constEnd() && (it.key() == sid) )
			{
				d->match.insertMulti(i, *it);
				++it;
			}
		}
	}
	
	//qDebug("reduction : %i => %i matches", match.count(), d->match.count());
	
	// determinize sub automatons
	
	foreach ( Automaton *sa, sub )
	{
		//qDebug("subcontext det...");
		if ( sa->det )
		{
			sa->ref();
			d->sub << sa;
		} else {
			Automaton *sd = sa->determinize();
			sd->parent = d;
			d->sub << sd;
		}
	}
	
	return d;
}

HashedAutomaton* Automaton::hashedDeterminize() const
{
	HashedAutomaton *d = new HashedAutomaton;
	d->id = id;
	d->det = true;
	d->flags = flags;
	d->format = format;
	
	int index = 0;
	QList<StateList> dstates;
	
	d->states.clear();
	d->standalone.clear();
	
	bool lastStandaloneFiller = false;
	//qDebug("det: 0x%x -> 0x%x", this, d);
	
	foreach ( StateId root, standalone )
	{
		if ( root == 0xffffffff )
		{
			// special filler state needed for recursive subcontexts
			d->standalone << 0xffffffff;
			lastStandaloneFiller = true;
			continue;
		} else if ( lastStandaloneFiller ) {
			if ( root == 0 )
				d->standalone << 0;
			else
				qWarning("unsupported use of std:filler");
			
			lastStandaloneFiller = false;
		}
		
		//qDebug("standalone state %i", n);
		StateId nroot = d->createStandaloneState();
		
		//qDebug("det(0x%x) : %i becomes %i [%i]", d, root, nroot, index);
		
		
		StateList l = epsilonClosure(root);
		
		if ( !l.contains(root) )
			l << root;
		
		//qDebug("epsilon-closure : %i elements", l.count());
		
		// sort states equiv to be able to use QList::operator == ()
		qSort(l);
		
		dstates << l;
		
		--index;
		
		while ( ++index < dstates.count() )
		{
			//qDebug() << "filling trans table for" << l;
			
			l = dstates.at(index);
			
			// compute local alphabet for DFA state
			QList<Transition> localAlphabet;
			
			foreach ( StateId sid, l )
			{
				const TransitionList& lt = trans.at(sid);
				
				foreach ( Transition t, lt )
				{
					if ( (t != Epsilon) && !localAlphabet.contains(t) )
						localAlphabet << t;
				}
			}
			
			// sort alphabet to place Any at end
			qSort(localAlphabet);
			
			//qDebug() << "alphabet" << localAlphabet;
			
			foreach ( Transition t, localAlphabet )
			{
				StateList cedges = epsilonClosure(move(l, t));
				
				qSort(cedges);
				
				int idx = dstates.indexOf(cedges);
				
				if ( idx == -1 )
				{
					// new state
					StateId ns = d->createState();
					
					if ( ns != (quint32)dstates.count() )
						qWarning("NFA=>DFA reduction error...");
					
					dstates << cedges;
					
					idx = ns;
				}
				
				d->addTransition(index, t, idx);
			}
		}
	}
	
	// backport matches
	
	for ( int i = 0; i < dstates.count(); ++i )
	{
		//
		const StateList& l = dstates.at(i);
		
		foreach ( StateId sid, l )
		{
			QHash<StateId, Match>::const_iterator it = match.constFind(sid);
			
			while ( it != match.constEnd() && (it.key() == sid) )
			{
				//qDebug("(0x%x, %i) => (0x%x, %i)", this, sid, d, i);
				d->match.insertMulti(i, *it);
				++it;
			}
		}
	}
	
	//qDebug("reduction : %i => %i matches", match.count(), d->match.count());
	
	// determinize sub automatons
	
	foreach ( Automaton *sa, sub )
	{
		//qDebug("subcontext det...");
		
		if ( sa == this )
		{
			d->sub << d;
		} else {
			HashedAutomaton *sd = sa->hashedDeterminize();
			sd->parent = d;
			d->sub << sd;
		}
	}
	
	// squeeze memory
	
	d->squeeze();
	
	return d;
}

/////////////////////////

HashedAutomaton::HashedAutomaton()
 : det(false), cref(0), format(0), flags(None), parent(0)
{
	ref();
	
	//qDebug("creating 0x%lx", this);
	
	createStandaloneState();
}

HashedAutomaton::~HashedAutomaton()
{
	if ( parent )
		parent->sub.removeAll(this);
	
	parent = 0;
	
	//qDeleteAll(sub);
	foreach ( HashedAutomaton *c, sub )
		c->deref();
}

void HashedAutomaton::ref()
{
	++cref;
}

void HashedAutomaton::deref()
{
	--cref;
	
	if ( cref == 0 )
		delete this;
}

void HashedAutomaton::squeeze()
{
	match.squeeze();
	
	for ( int i = 0; i < states.count(); ++i )
		states[i].squeeze();
	
	states.squeeze();
	
	foreach ( HashedAutomaton *c, sub )
		if ( c != this )
			c->squeeze();
}

QDataStream& operator << (QDataStream& s, const Match& m)
{
	s << m.type;
	s << m.format;
	
	return s;
}

QDataStream& operator >> (QDataStream& s, Match& m)
{
	s >> m.type;
	s >> m.format;
	
	return s;
}

HashedAutomaton* HashedAutomaton::deserialize(QDataStream& s)
{
	HashedAutomaton *a = new HashedAutomaton;
	
	s >> a->det;
	s >> a->id;
	s >> a->format;
	s >> a->flags;
	
	s >> a->states;
	s >> a->standalone;
	s >> a->match;
	
	int nchild = 0;
	
	s >> nchild;
	
	while ( nchild > 0 )
	{
		quint8 type;
		s >> type;
		
		if ( !type )
			a->sub << deserialize(s);
		else if ( type == 1 )
			a->sub << a;
		
		--nchild;
	}
	
	a->squeeze();
	
	return a;
}

void HashedAutomaton::serialize(HashedAutomaton *a, QDataStream& s)
{
	s << a->det;
	s << a->id;
	s << a->format;
	s << a->flags;
	
	s << a->states;
	s << a->standalone;
	s << a->match;
	
	s << a->sub.count();
	
	foreach ( HashedAutomaton *c, a->sub )
	{
		if ( c == a )
		{
			s << quint8(1);
		} else {
			s << quint8(0);
			serialize(c, s);
		}
	}
}

StateId HashedAutomaton::createState()
{
	StateId s = states.count();
	
	states << StateHash();
	
	return s;
}

StateId HashedAutomaton::createStandaloneState()
{
	StateId s = states.count();
	
	states << StateHash();
	
	standalone << s;
	
	return s;
}

void HashedAutomaton::addTransition(StateId from, Transition t, StateId to)
{
	//qDebug("%i == %c => %i", from, qTransitionChar(t), to);
	states[from][qTransitionChar(t)] = to;
}

HashedAutomaton* HashedAutomaton::clone(quint32 formatOffset) const
{
	
	return 0;
}
